package function.builtin;

import static function.FunctionTable.functionTable;

import java.text.MessageFormat;
import java.util.HashMap;

import error.LispException;
import function.*;
import function.builtin.cons.LIST;
import function.builtin.special.*;
import sexpression.*;

public class EVAL extends LispFunction {

    public static HashMap<String, LispFunction> getFunctionTable() {
        return functionTable;
    }

    public static LispFunction lookupFunctionOrLambda(SExpression functionExpression) {
        LispFunction function = lookupFunction(functionExpression.toString());

        if (function == null)
            function = createLambdaFunction(functionExpression);

        return function;
    }

    public static LispFunction lookupFunction(String functionName) {
        return functionTable.get(functionName);
    }

    private static LispFunction createLambdaFunction(SExpression lambdaExpression) {
        if (lambdaExpression.functionp())
            return ((LambdaExpression) lambdaExpression).getFunction();
        else if (LAMBDA.isLambdaExpression(lambdaExpression))
            return LAMBDA.createFunction((Cons) lambdaExpression);
        else
            throw new UndefinedFunctionException(lambdaExpression);
    }

    public static SExpression lookupSymbol(String symbolName) {
        if (symbolName.equals("NIL")) {
            return Nil.getInstance();
        } else if (symbolName.equals("T")) {
            return Symbol.T;
        } else if (symbolName.startsWith(":")) {
            return new Symbol(symbolName);
        }

        return SETF.lookup(symbolName);
    }

    public static boolean isDotted(Cons list) {
        if (list.nullp()) {
            return false;
        }

        SExpression cdr = list.getCdr();

        if (cdr.listp()) {
            return isDotted((Cons) cdr);
        }

        // the cdr of 'list' is not a list, therefore it is dotted
        return true;
    }

    public static SExpression eval(SExpression sExpression) {
        Cons argumentList = LIST.makeList(sExpression);
        EVAL eval = new EVAL();

        return eval.call(argumentList);
    }

    private ArgumentValidator argumentValidator;

    public EVAL() {
        this.argumentValidator = new ArgumentValidator("EVAL");
        this.argumentValidator.setExactNumberOfArguments(1);
    }

    public SExpression call(Cons argumentList) {
        argumentValidator.validate(argumentList);

        SExpression argument = argumentList.getCar();

        if (argument.listp()) {
            if (argument.consp())
                return evaluateList((Cons) argument);

            return argument; // NIL
        }

        if (argument.symbolp()) {
            SExpression symbolValue = lookupSymbol(argument.toString());

            if (symbolValue != null)
                return symbolValue;

            throw new UndefinedSymbolException(argument);
        }

        return argument; // NUMBER or STRING
    }

    private SExpression evaluateList(Cons list) {
        SExpression functionName = list.getCar();
        SExpression arguments = list.getCdr();
        LispFunction function = lookupFunctionOrLambda(functionName);

        ArgumentValidator functionListValidator = new ArgumentValidator(functionName.toString());
        functionListValidator.validate(list);

        Cons argumentList = (Cons) arguments;

        if (function.evaluateArguments())
            argumentList = evaluateArgList(argumentList);

        return function.call(argumentList);
    }

    private Cons evaluateArgList(Cons arguments) {
        if (arguments.nullp())
            return Nil.getInstance();

        SExpression car = eval(arguments.getCar());
        SExpression cdr = arguments.getCdr();

        return new Cons(car, evaluateArgList((Cons) cdr));
    }

    public static class UndefinedFunctionException extends LispException {

        private static final long serialVersionUID = 1L;
        private SExpression function;

        public UndefinedFunctionException(SExpression function) {
            this.function = function;
        }

        @Override
        public String getMessage() {
            return MessageFormat.format("undefined function: {0}", function);
        }
    }

    public static class UndefinedSymbolException extends LispException {

        private static final long serialVersionUID = 1L;
        private SExpression symbol;

        public UndefinedSymbolException(SExpression symbol) {
            this.symbol = symbol;
        }

        @Override
        public String getMessage() {
            return MessageFormat.format("symbol {0} has no value", symbol);
        }
    }

}
